#    This script is part of navis (http://www.github.com/schlegelp/navis).
#    Copyright (C) 2018 Philipp Schlegel
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

import csv
import functools
import json
import math
import numbers
import os
import uuid

import numpy as np
import scipy.spatial
import trimesh

from typing import Union, Optional, Sequence, List, Dict, Any
from typing_extensions import Literal

from .. import utils, config

# Set up logging
logger = config.logger


class Volume(trimesh.Trimesh):
    """Mesh consisting of vertices and faces.

    Subclass of ``trimesh.Trimesh`` with a few additional methods.

    Parameters
    ----------
    vertices :  list | array
                Vertices coordinates. Must be shape (N,3).
    faces :     list | array
                Indexed faceset.
    name :      str, optional
                Name of volume.
    color :     tuple, optional
                RGB color.
    id :        int, optional
                If not provided, neuron will be assigned a random UUID
                as ``.id``
    **kwargs
                Keyword arguments passed through to ``trimesh.Trimesh``

    See Also
    --------
    :func:`~navis.example_volume`
        Loads example volume(s).

    """

    def __init__(self,
                 vertices: Union[list, np.ndarray],
                 faces: Union[list, np.ndarray],
                 name: Optional[str] = None,
                 color: Union[str,
                              Sequence[Union[int, float]]] = (.95, .95, .95, .1),
                 id: Optional[int] = None, **kwargs):

        super().__init__(vertices=vertices, faces=faces, **kwargs)

        self.name: Optional[str] = name
        self.color: Union[str, Sequence[Union[int, float]]] = color
        self.id: Optional[int] = id if id else uuid.uuid4()

        # This is very hackish but we want to make sure that parent methods of
        # Trimesh return a navis.Volume instead of a trimesh.Trimesh
        for f in dir(trimesh.Trimesh):
            # Don't mess with magic/private methods
            if f.startswith('_'):
                continue
            # Skip properties
            if not callable(getattr(trimesh.Trimesh, f)):
                continue
            setattr(self, f, _force_volume(getattr(self, f)))

    @property
    def name(self):
        """Name of this volume."""
        return self.metadata.get('name')

    @name.setter
    def name(self, value):
        self.metadata['name'] = value

    @property
    def color(self):
        """Color used for plotting."""
        return self.metadata.get('color')

    @color.setter
    def color(self, value):
        self.metadata['color'] = value

    @property
    def id(self):
        """ID of this volume."""
        return self.metadata.get('id')

    @id.setter
    def id(self, value):
        self.metadata['id'] = value

    @classmethod
    def from_csv(cls,
                 vertices: str,
                 faces: str,
                 name: Optional[str] = None,
                 color: Union[str,
                              Sequence[Union[int, float]]] = (1, 1, 1, .1),
                 volume_id: Optional[int] = None, **kwargs) -> 'Volume':
        """Load volume from csv files containing vertices and faces.

        Parameters
        ----------
        vertices :      filepath | file-like
                        CSV file containing vertices.
        faces :         filepath | file-like
                        CSV file containing faces.
        **kwargs
                        Keyword arguments passed to ``csv.reader``.

        Returns
        -------
        navis.Volume

        """
        if not os.path.isfile(vertices) or not os.path.isfile(faces):
            raise ValueError('File(s) not found.')

        with open(vertices, 'r') as f:
            reader = csv.reader(f, **kwargs)
            vertices = np.array([r for r in reader]).astype(float)

        with open(faces, 'r') as f:
            reader = csv.reader(f, **kwargs)
            faces = np.array([r for r in reader]).astype(int)

        return cls(faces=faces, vertices=vertices, name=name, color=color,
                   volume_id=volume_id)

    def to_csv(self, filename: str, **kwargs) -> None:
        """Save volume as two separated csv files containing vertices and faces.

        Parameters
        ----------
        filename :      str
                        Filename to use. Will get a ``_vertices.csv`` and
                        ``_faces.csv`` suffix.
        **kwargs
                        Keyword arguments passed to ``csv.reader``.

        """
        for data, suffix in zip([self.faces, self.vertices],
                                ['_faces.csv', '_vertices.csv']):
            with open(filename + suffix, 'w') as csvfile:
                writer = csv.writer(csvfile)
                writer.writerows(data)

    @classmethod
    def from_json(cls,
                  filename: str,
                  name: Optional[str] = None,
                  color: Union[str,
                               Sequence[Union[int, float]]] = (1, 1, 1, .1),
                  import_kwargs: Dict = {},
                  **init_kwargs) -> 'Volume':
        """Load volume from json file containing vertices and faces.

        Parameters
        ----------
        filename
        import_kwargs
                        Keyword arguments passed to ``json.load``.
        **init_kwargs
                    Keyword arguments passed to navis.Volume upon
                    initialization.

        Returns
        -------
        navis.Volume

        """
        if not os.path.isfile(filename):
            raise ValueError('File not found.')

        with open(filename, 'r') as f:
            data = json.load(f, **import_kwargs)

        return cls(faces=data['faces'],
                   vertices=data['vertices'],
                   name=name, color=color, **init_kwargs)

    @classmethod
    def from_object(cls,
                    obj: Any,
                    name: Optional[str] = None,
                    color: Union[str,
                                 Sequence[Union[int, float]]] = (1, 1, 1, .1),
                    **init_kwargs) -> 'Volume':
        """Load volume from generic object that has ``.vertices`` and
        ``.faces`` attributes.

        Parameters
        ----------
        obj
        **init_kwargs
                    Keyword arguments passed to navis.Volume upon
                    initialization.

        Returns
        -------
        navis.Volume

        """
        if not hasattr(obj, 'vertices') or not hasattr(obj, 'faces'):
            raise ValueError('Object must have faces and vertices attributes.')

        return cls(faces=obj.faces, vertices=obj.vertices,
                   name=name, color=color, **init_kwargs)

    @classmethod
    def from_file(cls,
                  filename: str,
                  name: Optional[str] = None,
                  color: Union[str,
                               Sequence[Union[int, float]]] = (1, 1, 1, .1),
                  import_kwargs: Dict = {},
                  **init_kwargs) -> 'Volume':
        """Load volume from file.

        Parameters
        ----------
        filename :      str
                        File to load from.
        import_kwargs
                        Keyword arguments passed to importer:
                          - ``json.load`` for JSON file
                          - ``trimesh.load_mesh`` for OBJ and STL files
        **init_kwargs
                    Keyword arguments passed to navis.Volume upon
                    initialization.

        Returns
        -------
        navis.Volume

        """
        if not os.path.isfile(filename):
            raise ValueError('File not found.')

        f, ext = os.path.splitext(filename)

        if ext == '.json':
            return cls.from_json(filename=filename, name=name, color=color,
                                 import_kwargs=import_kwargs, **init_kwargs)

        try:
            import trimesh
        except ImportError:
            raise ImportError('Unable to import: trimesh missing - please '
                              'install: "pip install trimesh"')
        except BaseException:
            raise

        tm = trimesh.load_mesh(filename, **import_kwargs)

        return cls.from_object(tm, name=name, color=color, **init_kwargs)

    def to_json(self, filename: str) -> None:
        """Save volume as json file.

        Parameters
        ----------
        filename :      str
                        Filename to use.
        """
        with open(filename, 'w') as f:
            json.dump({'vertices': self.vertices.tolist(),
                       'faces': self.faces.tolist()},
                      f)

    @classmethod
    def combine(cls,
                x: Sequence['Volume'],
                name: str = 'comb_vol',
                color: Union[str,
                             Sequence[Union[int, float]]] = (1, 1, 1, .1)
                ) -> 'Volume':
        """Merge multiple volumes into a single object.

        Parameters
        ----------
        x :     list or dict of Volumes
        name :  str, optional
                Name of the combined volume.
        color : tuple | str, optional
                Color of the combined volume.

        Returns
        -------
        :class:`~navis.Volume`

        """
        if isinstance(x, Volume):
            return x

        if isinstance(x, dict):
            x = list(x.values())

        if not utils.is_iterable(x):
            x = [x]  # type: ignore

        if False in [isinstance(v, Volume) for v in x]:
            raise TypeError('Input must be list of volumes')

        vertices: np.ndarray = np.empty((0, 3))
        faces: List[List[int]] = []

        # Reindex faces
        for vol in x:
            offs = len(vertices)
            vertices = np.append(vertices, vol.vertices, axis=0)
            faces += [[f[0] + offs, f[1] + offs, f[2] + offs]
                      for f in vol.faces]

        return cls(vertices=vertices, faces=faces, name=name, color=color)

    @property
    def bbox(self) -> np.ndarray:
        """Bounding box of this volume. """
        return self.bounds

    @property
    def verts(self) -> np.ndarray:
        """Legacy access to ``.vertices``."""
        return self.vertices

    @verts.setter
    def verts(self, v):
        self.vertices = v

    @property
    def center(self) -> np.ndarray:
        """Center of mass."""
        return np.mean(self.vertices, axis=0)

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        """
        Print quick summary of the current geometry without
        computing properties.
        """
        elements = []
        if hasattr(self, 'name'):
            # for Trimesh
            elements.append(f'name={self.name}')
        if hasattr(self, 'id') and not isinstance(self.id, uuid.UUID):
            # for Trimesh
            elements.append(f'id={self.id}')
        if hasattr(self, 'color'):
            # for Trimesh
            elements.append(f'color={self.color}')
        if hasattr(self, 'vertices'):
            # for Trimesh and PointCloud
            elements.append(f'vertices.shape={self.vertices.shape}')
        if hasattr(self, 'faces'):
            # for Trimesh
            elements.append(f'faces.shape={self.faces.shape}')
        return f'<navis.Volume({", ".join(elements)})>'

    def __truediv__(self, other):
        """Implement division for vertex coordinates."""
        if isinstance(other, numbers.Number):
            # If a number, consider this an offset for coordinates
            return self.__mul__(1 / other)
        else:
            return NotImplemented

    def __mul__(self, other):
        """Implement multiplication for vertex coordinates."""
        if isinstance(other, numbers.Number):
            # If a number, consider this an offset for coordinates
            v = self.copy()
            v.vertices *= other
            return v
        else:
            return NotImplemented

    def resize(self,
               x: Union[float, int],
               method: Union[Literal['center'],
                             Literal['centroid'],
                             Literal['normals'],
                             Literal['origin']] = 'center',
               inplace: bool = True) -> Optional['Volume']:
        """Resize volume.

        Parameters
        ----------
        x :         int | float
                    Resizing factor. For methods "center", "centroid" and
                    "origin" this is the fraction of original size (e.g.
                    ``.5`` for half size). For method "normals", this is
                    is the absolute displacement (e.g. ``-1000`` to shrink
                    volume by that many units)!
        method :    "center" | "centroid" | "normals" | "origin"
                    Point in space to use for resizing.

                    .. list-table::
                        :widths: 15 75
                        :header-rows: 1

                        * - method
                          - explanation
                        * - center
                          - average of all vertices
                        * - centroid
                          - average of the triangle centroids weighted by the
                            area of each triangle.
                        * - origin
                          - resizes relative to origin of coordinate system
                            (0, 0, 0)
                        * - normals
                          - resize using face normals. Note that this method
                            uses absolute displacement for parameter ``x``.

        inplace :   bool, optional
                    If False, will return resized copy.

        Returns
        -------
        :class:`navis.Volume`
                    Resized copy of original volume. Only if ``inplace=False``.
        None
                    If ``inplace=True``.

        """
        assert isinstance(method, str)

        method = method.lower()

        perm_methods = ['center', 'origin', 'normals', 'centroid']
        if method not in perm_methods:
            raise ValueError('Unknown method "{}". Allowed '
                             'methods: {}'.format(method,
                                                  ', '.join(perm_methods)))

        if not inplace:
            v = self.copy()
        else:
            v = self

        if method == 'normals':
            v.vertices = v.vertices + (v.vertex_normals * x)
        else:
            # Get the center
            if method == 'center':
                cn = np.mean(v.vertices, axis=0)
            elif method == 'centroid':
                cn = v.to_trimesh().centroid
            elif method == 'origin':
                cn = np.array([0, 0, 0])

            # Get vector from center to each vertex
            vec = v.vertices - cn

            # Multiply vector by resize factor
            vec *= x

            # Recalculate vertex positions
            v.vertices = vec + cn

        # Make sure to reset any pyoctree data on this volume
        if hasattr(v, 'pyoctree'):
            delattr(v, 'pyoctree')

        if not inplace:
            return v

    def plot3d(self, **kwargs):
        """Plot volume using :func:`navis.plot3d`.

        Parameters
        ----------
        **kwargs
                Keyword arguments. Will be passed to :func:`navis.plot3d`.
                See ``help(navis.plot3d)`` for a list of keywords.

        See Also
        --------
        :func:`navis.plot3d`
                    Function called to generate 3d plot.

        Examples
        --------
        >>> vol = navis.get_volume('v14.LH_R')
        >>> vol.plot3d(color = (255, 0, 0))
        """

        from .. import plotting

        if 'color' in kwargs:
            self.color = kwargs['color']

        return plotting.plot3d(self, **kwargs)

    def show(self, **kwargs):
        """See ``.plot3d``."""
        return self.plot3d(**kwargs)

    def _outlines_3d(self, view='xy', **kwargs):
        """Generate 3d outlines along a given view (see ``.to_2d()``).

        Parameters
        ----------
        **kwargs
                    Keyword arguments passed to :func:`~navis.Volume.to_2d`.

        Returns
        -------
        list
                    Coordinates of 2d circumference.
                    e.g. ``[(x1, y1, z1), (x2, y2, z2), (x3, y3, z3), ...]``
                    Third dimension is averaged.

        """
        co2d = np.array(self.to_2d(view=view, **kwargs))

        if view in ['xy', 'yx']:
            third = np.repeat(self.center[2], co2d.shape[0])
        elif view in ['xz', 'zx']:
            third = np.repeat(self.center[1], co2d.shape[0])
        elif view in ['yz', 'zy']:
            third = np.repeat(self.center[0], co2d.shape[0])

        return np.append(co2d, third.reshape(co2d.shape[0], 1), axis=1)

    def to_2d(self,
              alpha: float = 0.00017,
              view: str = 'xy',
              invert_y: bool = False) -> Sequence[Union[float, int]]:
        """Compute the 2d alpha shape (concave hull) this volume.

        Uses Scipy Delaunay and shapely.

        Parameters
        ----------
        alpha:      float, optional
                    Alpha value to influence the gooeyness of the border.
                    Smaller numbers don't fall inward as much as larger
                    numbers. Too large, and you lose everything!
        view :      'xy' | 'xz' | 'yz', optional
                    Determines if frontal, lateral or top view.

        Returns
        -------
        list
                    Coordinates of 2d circumference
                    e.g. ``[(x1, y1), (x2, y2), (x3, y3), ...]``

        """
        def add_edge(edges, edge_points, coords, i, j):
            """Add line between the i-th and j-th points."""
            if (i, j) in edges or (j, i) in edges:
                # already added
                return
            edges.add((i, j))
            edge_points.append(coords[[i, j]])

        accepted_views = ['xy', 'xz', 'yz']

        try:
            from shapely.ops import cascaded_union, polygonize  # type: ignore
            import shapely.geometry as geometry  # type: ignore
        except ImportError:
            raise ImportError('This function needs the shapely package.')

        coords: np.ndarray

        if view in['xy', 'yx']:
            coords = self.vertices[:, [0, 1]]  # type: ignore
            if invert_y:
                coords[:, 1] = coords[:, 1] * -1  # type: ignore

        elif view in ['xz', 'zx']:
            coords = self.vertices[:, [0, 2]]  # type: ignore
        elif view in ['yz', 'zy']:
            coords = self.vertices[:, [1, 2]]  # type: ignore
            if invert_y:
                coords[:, 0] = coords[:, 0] * -1  # type: ignore
        else:
            raise ValueError(f'View {view} unknown. Please use either: {accepted_views}')

        tri = scipy.spatial.Delaunay(coords)
        edges: set = set()
        edge_points: list = []
        # loop over triangles:
        # ia, ib, ic = indices of corner points of the
        # triangle
        for ia, ib, ic in tri.vertices:
            pa: np.ndarray = coords[ia]  # type: ignore
            pb: np.ndarray = coords[ib]  # type: ignore
            pc: np.ndarray = coords[ic]  # type: ignore
            # Lengths of sides of triangle
            a = math.sqrt((pa[0] - pb[0])**2 + (pa[1] - pb[1])**2)  # type: ignore
            b = math.sqrt((pb[0] - pc[0])**2 + (pb[1] - pc[1])**2)  # type: ignore
            c = math.sqrt((pc[0] - pa[0])**2 + (pc[1] - pa[1])**2)  # type: ignore
            # Semiperimeter of triangle
            s = (a + b + c) / 2.0
            # Area of triangle by Heron's formula
            area = math.sqrt(s * (s - a) * (s - b) * (s - c))
            circum_r = a * b * c / (4.0 * area)
            # Here's the radius filter.
            if circum_r < 1.0 / alpha:
                add_edge(edges, edge_points, coords, ia, ib)
                add_edge(edges, edge_points, coords, ib, ic)
                add_edge(edges, edge_points, coords, ic, ia)

        m = geometry.MultiLineString(edge_points)
        triangles = list(polygonize(m))
        concave_hull = cascaded_union(triangles)

        # Try with current settings, if this is not successful, try again
        # with lower alpha
        try:
            return list(concave_hull.exterior.coords)
        except AttributeError:
            return self.to_2d(alpha=alpha / 10, view=view, invert_y=invert_y)
        except BaseException:
            raise


def _force_volume(f):
    """Convert result from ``trimesh.Trimesh`` to ``navis.Volume``."""
    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        res = f(*args, **kwargs)
        if isinstance(res, trimesh.Trimesh):
            res = Volume(res.vertices, res.faces)
        return res
    return wrapper



#    This script is part of pymaid (http://www.github.com/schlegelp/navis).
#    but has been adapted from vispy (http://www.vispy.org)

import collections
import warnings

with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    from vispy.visuals import MeshVisual
    from vispy.color import ColorArray

    from vispy.scene.visuals import create_visual_node
    from vispy.util.transforms import rotate

import numpy as np
from numpy.linalg import norm


class NeuronVisual(MeshVisual):
    """Displays a tube around a piecewise-linear path.
    The tube mesh is corrected following its Frenet curvature and
    torsion such that it varies smoothly along the curve, including if
    the tube is closed.

    Parameters
    ----------
    points : list of arrays
        An list of arrays of (x, y, z) points describing the path along which
        a tube will be extruded.
    radius : list of floats
        The radius of the tube. Defaults to 1.0.
    closed : bool
        Whether the tube should be closed, joining the last point to the
        first. Defaults to False.
    color : Color | ColorArray
        The color(s) to use when drawing the tube. The same color is
        applied to each vertex of the mesh surrounding each point of
        the line. If the input is a ColorArray, the argument will be
        cycled; for instance if 'red' is passed then the entire tube
        will be red, or if ['green', 'blue'] is passed then the points
        will alternate between these colours. Defaults to 'purple'.
    tube_points : int
        The number of points in the circle-approximating polygon of the
        tube's cross section. Defaults to 8.
    shading : str | None
        Same as for the `MeshVisual` class. Defaults to 'smooth'.
    vertex_colors: ndarray | None
        Same as for the `MeshVisual` class.
    face_colors: ndarray | None
        Same as for the `MeshVisual` class.
    mode : str
        Same as for the `MeshVisual` class. Defaults to 'triangles'.
    """

    def __init__(self, segments, radii=1.0,
                 color='purple',
                 tube_points=8,
                 shading='smooth',
                 vertex_colors=None,
                 face_colors=None,
                 use_normals=True,
                 mode='triangles'):

        vertices = np.empty((0, 3), dtype=np.float)
        indices = np.empty((0, 3), dtype=np.uint32)

        if not isinstance(radii, collections.Iterable):
                radii = [[radii] * len(points) for points in segments]

        for points, radius in zip(segments, radii):
            # Need to make sure points are floats
            points = np.array(points).astype(float)

            if use_normals:
                tangents, normals, binormals = _frenet_frames(points)
            else:
                tangents = normals = binormals = np.ones((len(points), 3))

            n_segments = len(points) - 1

            if not isinstance(radius, collections.Iterable):
                radius = [radius] * len(points)

            radius = np.array(radius)

            # Vertices for each point on the circle
            verts = np.repeat(points, tube_points, axis=0)

            v = np.arange(tube_points,
                          dtype=np.float) / tube_points * 2 * np.pi

            all_cx = (radius * -1. * np.tile(np.cos(v), points.shape[0]).reshape((tube_points, points.shape[0]), order='F')).T
            cx_norm = (all_cx[:, :, np.newaxis] * normals[:, np.newaxis, :]).reshape(verts.shape)

            all_cy = (radius * np.tile(np.sin(v), points.shape[0]).reshape((tube_points, points.shape[0]), order='F')).T
            cy_norm = (all_cy[:, :, np.newaxis] * binormals[:, np.newaxis, :]).reshape(verts.shape)

            verts = verts + cx_norm + cy_norm

            """
            # get the positions of each vertex
            grid = np.zeros((len(points), tube_points, 3))

            for i in range(len(points)):
                pos = points[i]
                if use_normals:
                    normal = normals[i]
                    binormal = binormals[i]
                else:
                    normal = binormal = 1

                r = radius[i]

                # Add a vertex for each point on the circle
                v = np.arange(tube_points,
                              dtype=np.float) / tube_points * 2 * np.pi
                cx = -1. * r * np.cos(v)
                cy = r * np.sin(v)
                grid[i] = (pos + cx[:, np.newaxis] * normal +
                           cy[:, np.newaxis] * binormal)
            """

            # Generate indices for the first segment
            ix = np.arange(0, tube_points)

            # Repeat indices n_segments-times
            ix = np.tile(ix, n_segments)

            # Offset indices by number segments and tube points
            offsets = np.repeat((np.arange(0, n_segments)) * tube_points, tube_points)
            ix += offsets

            # Turn indices into faces
            ix_a = ix
            ix_b = ix + tube_points

            ix_c = ix_b.reshape((n_segments, tube_points))
            ix_c = np.append(ix_c[:, 1:], ix_c[:, [0]], axis=1)
            ix_c = ix_c.ravel()

            ix_d = ix_a.reshape((n_segments, tube_points))
            ix_d = np.append(ix_d[:, 1:], ix_d[:, [0]], axis=1)
            ix_d = ix_d.ravel()

            faces1 = np.concatenate((ix_a, ix_b, ix_d), axis=0).reshape((n_segments * tube_points, 3), order='F')
            faces2 = np.concatenate((ix_b, ix_c, ix_d), axis=0).reshape((n_segments * tube_points, 3), order='F')

            faces = np.append(faces1, faces2, axis=0)

            # Offset faces against already existing vertices
            faces += vertices.shape[0]

            # Get vertices from grid
            #this_vertices = grid.reshape(grid.shape[0] * grid.shape[1], 3)
            this_vertices = verts

            # Add vertices and faces to total collection
            vertices = np.append(vertices, this_vertices, axis=0)
            indices = np.append(indices, faces, axis=0)

        color = ColorArray(color)
        if vertex_colors is None:
            vertex_colors = np.resize(color.rgba,
                                     (vertices.shape[0], 4))

        MeshVisual.__init__(self, vertices, indices,
                            vertex_colors=vertex_colors,
                            face_colors=face_colors,
                            shading=shading,
                            mode=mode)


class TubeVisual(MeshVisual):
    """Displays a tube around a piecewise-linear path.
    The tube mesh is corrected following its Frenet curvature and
    torsion such that it varies smoothly along the curve, including if
    the tube is closed.
    Parameters
    ----------
    points : ndarray
        An array of (x, y, z) points describing the path along which the
        tube will be extruded.
    radius : float
        The radius of the tube. Defaults to 1.0.
    closed : bool
        Whether the tube should be closed, joining the last point to the
        first. Defaults to False.
    color : Color | ColorArray
        The color(s) to use when drawing the tube. The same color is
        applied to each vertex of the mesh surrounding each point of
        the line. If the input is a ColorArray, the argument will be
        cycled; for instance if 'red' is passed then the entire tube
        will be red, or if ['green', 'blue'] is passed then the points
        will alternate between these colours. Defaults to 'purple'.
    tube_points : int
        The number of points in the circle-approximating polygon of the
        tube's cross section. Defaults to 8.
    shading : str | None
        Same as for the `MeshVisual` class. Defaults to 'smooth'.
    vertex_colors: ndarray | None
        Same as for the `MeshVisual` class.
    face_colors: ndarray | None
        Same as for the `MeshVisual` class.
    mode : str
        Same as for the `MeshVisual` class. Defaults to 'triangles'.
    """

    def __init__(self, points, radius=1.0,
                 closed=False,
                 color='purple',
                 tube_points=8,
                 shading='smooth',
                 vertex_colors=None,
                 face_colors=None,
                 mode='triangles'):

        points = np.array(points)

        tangents, normals, binormals = _frenet_frames(points, closed)

        segments = len(points) - 1

        if not isinstance(radius, collections.Iterable):
            radius = [radius] * len(points)

        # get the positions of each vertex
        grid = np.zeros((len(points), tube_points, 3))
        for i in range(len(points)):
            pos = points[i]
            normal = normals[i]
            binormal = binormals[i]
            r = radius[i]

            # Add a vertex for each point on the circle
            v = np.arange(tube_points,
                          dtype=np.float) / tube_points * 2 * np.pi
            cx = -1. * r * np.cos(v)
            cy = r * np.sin(v)
            grid[i] = (pos + cx[:, np.newaxis] * normal +
                       cy[:, np.newaxis] * binormal)

        # construct the mesh
        indices = []
        for i in range(segments):
            for j in range(tube_points):
                ip = (i + 1) % segments if closed else i + 1
                jp = (j + 1) % tube_points

                index_a = i * tube_points + j
                index_b = ip * tube_points + j
                index_c = ip * tube_points + jp
                index_d = i * tube_points + jp

                indices.append([index_a, index_b, index_d])
                indices.append([index_b, index_c, index_d])

        vertices = grid.reshape(grid.shape[0] * grid.shape[1], 3)

        color = ColorArray(color)
        if vertex_colors is None:
            point_colors = np.resize(color.rgba,
                                     (len(points), 4))
            vertex_colors = np.repeat(point_colors, tube_points, axis=0)

        indices = np.array(indices, dtype=np.uint32)

        MeshVisual.__init__(self, vertices, indices,
                            vertex_colors=vertex_colors,
                            face_colors=face_colors,
                            shading=shading,
                            mode=mode)


def _frenet_frames(points):
    '''Calculates and returns the tangents, normals and binormals for
    the tube.'''
    tangents = np.zeros((len(points), 3))
    normals = np.zeros((len(points), 3))

    epsilon = 0.0001

    # Compute tangent vectors for each segment
    tangents = np.roll(points, -1, axis=0) - np.roll(points, 1, axis=0)

    tangents[0] = points[1] - points[0]
    tangents[-1] = points[-1] - points[-2]

    mags = np.sqrt(np.sum(tangents * tangents, axis=1))
    tangents /= mags[:, np.newaxis]

    # Get initial normal and binormal
    t = np.abs(tangents[0])

    smallest = np.argmin(t)
    normal = np.zeros(3)
    normal[smallest] = 1.

    vec = np.cross(tangents[0], normal)
    normals[0] = np.cross(tangents[0], vec)

    all_vec = np.cross(tangents[:-1], tangents[1:])
    all_vec_norm = norm(all_vec, axis=1)

    # Normalise vectors if necessary
    where = all_vec_norm > epsilon
    all_vec[where, :] /=  all_vec_norm[where].reshape((sum(where), 1))

    # Precompute inner dot product
    dp = np.sum(tangents[:-1] * tangents[1:], axis=1)
    # Clip
    cl = np.clip(dp, -1, 1)
    # Get theta
    th = np.arccos(cl)

    # Compute normal and binormal vectors along the path
    for i in range(1, len(points)):
        normals[i] = normals[i-1]

        vec_norm = all_vec_norm[i-1]
        vec = all_vec[i-1]
        if vec_norm > epsilon:
            normals[i] = rotate(-np.degrees(th[i-1]),
                                vec)[:3, :3].dot(normals[i])

    binormals = np.cross(tangents, normals)

    return tangents, normals, binormals


Tube = create_visual_node(TubeVisual)
MeshNeuron = create_visual_node(NeuronVisual)



from setuptools import setup, find_packages

import re

VERSIONFILE = "navis/__version__.py"
verstrline = open(VERSIONFILE, "rt").read()
VSRE = r"^__version__ = ['\"]([^'\"]*)['\"]"
mo = re.search(VSRE, verstrline, re.M)
if mo:
    verstr = mo.group(1)
else:
    raise RuntimeError("Unable to find version string in %s." % (VERSIONFILE,))

with open('requirements.txt') as f:
    requirements = f.read().splitlines()
    requirements = [l for l in requirements if not l.startswith('#')]

LONG_DESCRIPTION = """
NAVis is a Python 3 library for Neuron Analysis and
Visualization with focus on hierarchical tree-like neuron data.

Features:

  - interactive 2D (matplotlib) and 3D (vispy/plotly) plotting of neurons
  - virtual neuron surgery: cutting, pruning, rerooting
  - clustering methods (e.g. by connectivity or synapse placement)
  - R bindings (e.g. for libraries nat, rcatmaid, elmr)
  - interfaces with Blender 3D and Cytoscape

Check out the `Documentation <http://navis.readthedocs.io/>`_.
"""

setup(
    name='navis',
    version=verstr,
    packages=find_packages(),
    license='GNU GPL V3',
    description='Neuron Analysis and Visualization Library',
    long_description=LONG_DESCRIPTION,
    url='http://navis.readthedocs.io',
    author='Philipp Schlegel',
    author_email='pms70@cam.ac.uk',
    keywords='Neuron Analysis Visualization Anatomy Connectivity',
    classifiers=[
        'Development Status :: 3 - Alpha',

        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Bio-Informatics',

        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',

        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    install_requires=requirements,
    extras_require={'extras': ['pyoctree~=0.2.10']},
    python_requires='>=3.6',
    zip_safe=False,

    include_package_data=True

)



#    This script is part of navis (http://www.github.com/schlegelp/navis).
#    but has been adapted from vispy (http://www.vispy.org)
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

from vispy.visuals import MeshVisual
from vispy.color import ColorArray
import numpy as np
from vispy.visuals.tube import _frenet_frames
from vispy.scene.visuals import create_visual_node

import collections


class TubeVisual(MeshVisual):
    """Displays a tube around a piecewise-linear path.
    The tube mesh is corrected following its Frenet curvature and
    torsion such that it varies smoothly along the curve, including if
    the tube is closed.

    Parameters
    ----------
    points : ndarray
        An array of (x, y, z) points describing the path along which the
        tube will be extruded.
    radius : float
        The radius of the tube. Defaults to 1.0.
    closed : bool
        Whether the tube should be closed, joining the last point to the
        first. Defaults to False.
    color : Color | ColorArray
        The color(s) to use when drawing the tube. The same color is
        applied to each vertex of the mesh surrounding each point of
        the line. If the input is a ColorArray, the argument will be
        cycled; for instance if 'red' is passed then the entire tube
        will be red, or if ['green', 'blue'] is passed then the points
        will alternate between these colours. Defaults to 'purple'.
    tube_points : int
        The number of points in the circle-approximating polygon of the
        tube's cross section. Defaults to 8.
    shading : str | None
        Same as for the `MeshVisual` class. Defaults to 'smooth'.
    vertex_colors: ndarray | None
        Same as for the `MeshVisual` class.
    face_colors: ndarray | None
        Same as for the `MeshVisual` class.
    mode : str
        Same as for the `MeshVisual` class. Defaults to 'triangles'.
    """

    def __init__(self, points, radius=1.0,
                 closed=False,
                 color='purple',
                 tube_points=8,
                 shading='smooth',
                 vertex_colors=None,
                 face_colors=None,
                 mode='triangles'):

        points = np.array(points)

        tangents, normals, binormals = _frenet_frames(points, closed)

        segments = len(points) - 1

        if not isinstance(radius, collections.Iterable):
            radius = [radius] * len(points)

        # get the positions of each vertex
        grid = np.zeros((len(points), tube_points, 3))
        for i in range(len(points)):
            pos = points[i]
            normal = normals[i]
            binormal = binormals[i]
            r = radius[i]

            # Add a vertex for each point on the circle
            v = np.arange(tube_points,
                          dtype=np.float) / tube_points * 2 * np.pi
            cx = -1. * r * np.cos(v)
            cy = r * np.sin(v)
            grid[i] = (pos + cx[:, np.newaxis] * normal +
                       cy[:, np.newaxis] * binormal)

        # construct the mesh
        indices = []
        for i in range(segments):
            for j in range(tube_points):
                ip = (i + 1) % segments if closed else i + 1
                jp = (j + 1) % tube_points

                index_a = i * tube_points + j
                index_b = ip * tube_points + j
                index_c = ip * tube_points + jp
                index_d = i * tube_points + jp

                indices.append([index_a, index_b, index_d])
                indices.append([index_b, index_c, index_d])

        vertices = grid.reshape(grid.shape[0] * grid.shape[1], 3)

        color = ColorArray(color)
        if vertex_colors is None:
            point_colors = np.resize(color.rgba,
                                     (len(points), 4))
            vertex_colors = np.repeat(point_colors, tube_points, axis=0)

        indices = np.array(indices, dtype=np.uint32)

        MeshVisual.__init__(self, vertices, indices,
                            vertex_colors=vertex_colors,
                            face_colors=face_colors,
                            shading=shading,
                            mode=mode)


Tube = create_visual_node(TubeVisual)



#    This script is part of navis (http://www.github.com/schlegelp/navis).
#    Copyright (C) 2018 Philipp Schlegel
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

import numpy as np
import pandas as pd

import plotly.graph_objs as go

from ..colors import *
from ..plot_utils import *
from ... import core, utils

__all__ = ['neuron2plotly', 'scatter2plotly', 'dotprops2plotly',
           'volume2plotly', 'layout2plotly']

# Generate sphere for somas
fib_points = fibonacci_sphere(samples=30)


def neuron2plotly(x, **kwargs):
    """ Converts TreeNeurons to plotly objects."""

    if isinstance(x, core.TreeNeuron):
        x = core.NeuronList(x)
    elif isinstance(x, core.NeuronList):
        pass
    else:
        raise TypeError('Unable to process data of type "{}"'.format(type(x)))

    colors = kwargs.get('color',
                        kwargs.get('c',
                                   kwargs.get('colors', None)))

    colormap, _, _ = prepare_colormap(colors,
                                      x, None,
                                      use_neuron_color=kwargs.get('use_neuron_color', False),
                                      color_range=255)

    linewidth = kwargs.get('linewidth', 1)

    syn_lay = {
        0: {
            'name': 'Presynapses',
            'color': (1, 0, 0)
        },
        1: {
            'name': 'Postsynapses',
            'color': (0, 0, 1)
        },
        2: {
            'name': 'Gap junctions',
            'color': (0, 1, 0)
        },
        'display': 'lines'  # 'circle'
    }
    syn_lay.update(kwargs.get('synapse_layout', {}))

    trace_data = []
    for i, neuron in enumerate(x):
        color = colormap[i]
        name = str(getattr(neuron, 'name', None))

        if not kwargs.get('connectors_only', False):
            coords = segments_to_coords(neuron, neuron.segments)

            # We have to add (None, None, None) to the end of each slab to
            # make that line discontinuous there
            coords = np.vstack([np.append(t, [[None] * 3], axis=0) for t in coords])

            if kwargs.get('by_strahler', False):
                s_index = morpho.strahler_index(neuron, return_dict=True)
                c = []
                for k, s in enumerate(coords):
                    this_c = 'rgba(%i,%i,%i,%f)' % (color[0],
                                                    color[1],
                                                    color[2],
                                                    s_index[s[0]] / max(s_index.values()))
                    # Slabs are separated by a <None> coordinate -> this is
                    # why we need one more color entry
                    c += [this_c] * (len(s) + 1)
            else:
                try:
                    c = 'rgb{}'.format(color)
                except BaseException:
                    c = 'rgb(10,10,10)'

            trace_data.append(go.Scatter3d(x=coords[:, 0],
                                           y=coords[:, 1],
                                           z=coords[:, 2],
                                           mode='lines',
                                           line=dict(color=c,
                                                     width=linewidth),
                                           name=name,
                                           legendgroup=name,
                                           showlegend=True,
                                           hoverinfo='none'
                                           ))

            # Add soma(s):
            soma = utils.make_iterable(neuron.soma)
            if any(soma):
                for s in soma:
                    n = neuron.nodes.set_index('node_id').loc[s]
                    r = getattr(n, neuron.soma_radius) if isinstance(neuron.soma_radius, str) else neuron.soma_radius
                    try:
                        c = 'rgb{}'.format(color)
                    except BaseException:
                        c = 'rgb(10,10,10)'
                    trace_data.append(go.Mesh3d(
                        x=[(v[0] * r) + n.x for v in fib_points],
                        # y and z are switched
                        y=[(v[1] * r) + n.y for v in fib_points],
                        z=[(v[2] * r) + n.z for v in fib_points],

                        alphahull=.5,
                        color=c,
                        name=name,
                        hoverinfo='name'))

        # Add connectors
        if kwargs.get('connectors', False) or \
           kwargs.get('connectors_only', False):
            for j in neuron.connectors.type.unique():
                if kwargs.get('cn_mesh_colors', False):
                    c = color
                else:
                    c = syn_lay.get(j, {'color': (10, 10, 10)})['color']

                this_cn = neuron.connectors[neuron.connectors.type == j]

                if syn_lay['display'] == 'circles':
                    trace_data.append(go.Scatter3d(
                        x=this_cn.x.values,
                        y=this_cn.zy.values,
                        z=this_cn.z.values,
                        mode='markers',
                        marker=dict(color='rgb%s' % str(c), size=2),
                        name=syn_lay.get(j, {'name': 'connector'})['name'] + ' of ' + name,
                        showlegend=True,
                        hoverinfo='none'
                    ))
                elif syn_lay['display'] == 'lines':
                    # Find associated treenode
                    tn = neuron.nodes.set_index('node_id').loc[this_cn.node_id.values]
                    x_coords = [n for sublist in zip(this_cn.x.values, tn.x.values, [None] * this_cn.shape[0]) for n in sublist]
                    y_coords = [n for sublist in zip(this_cn.y.values, tn.y.values, [None] * this_cn.shape[0]) for n in sublist]
                    z_coords = [n for sublist in zip(this_cn.z.values, tn.z.values, [None] * this_cn.shape[0]) for n in sublist]

                    trace_data.append(go.Scatter3d(
                        x=x_coords,
                        y=y_coords,
                        z=z_coords,
                        mode='lines',
                        line=dict(
                            color='rgb%s' % str(c),
                            width=5
                        ),
                        name=syn_lay.get(j, {'name': 'connector'})['name'] + ' of ' + name,
                        showlegend=True,
                        hoverinfo='none'
                    ))
                else:
                    raise ValueError('Unknown display type for connectors "{}"'.format(syn_lay['display']))

    return trace_data


def scatter2plotly(x, **kwargs):
    """ Converts DataFrame with x,y,z columns to plotly scatter plot."""

    c = eval_color(kwargs.get('color', (10, 10, 10)), color_range=255)
    s = kwargs.get('size', 2)
    name = kwargs.get('name', None)

    trace_data = []
    for scatter in x:
        if isinstance(scatter, pd.DataFrame):
            if not all([c in scatter.columns for c in ['x', 'y', 'z']]):
                raise ValueError('DataFrame must have x, y and z columns')
            scatter = [['x', 'y', 'z']].values

        if not isinstance(scatter, np.ndarray):
            scatter = np.array(scatter)

        trace_data.append(go.Scatter3d(x=scatter[:, 0],
                                       y=scatter[:, 1],
                                       z=scatter[:, 2],
                                       mode='markers',
                                       marker=dict(color='rgb%s' % str(c),
                                                   size=s,
                                                   opacity=kwargs.get('opacity', 1)),
                                       name=name,
                                       showlegend=True,
                                       hoverinfo='none'))
    return trace_data


def lines2plotly(x, **kwargs):
    """Convert DataFrame with x, y, z, x1, y1, z1 columns to line plots."""
    name = kwargs.get('name', None)
    c = kwargs.get('color', (10, 10, 10))

    x_coords = [n for sublist in zip(x.x.values, x.x1.values,
                                     [None] * x.shape[0]) for n in sublist]
    y_coords = [n for sublist in zip(x.y.values, x.y1.values,
                                     [None] * x.shape[0]) for n in sublist]
    z_coords = [n for sublist in zip(x.z.values, x.z1.values,
                                     [None] * x.shape[0]) for n in sublist]

    trace_data = []
    trace_data.append(go.Scatter3d(x=x_coords,
                                   y=y_coords,
                                   z=z_coords,
                                   mode='lines',
                                   line=dict(
                                        color=f'rgb{str(c)}',
                                        width=5
                                   ),
                                   name=name,
                                   showlegend=True,
                                   hoverinfo='none'))

    return trace_data


def dotprops2plotly(x, **kwargs):
    """ Converts dotprops to plotly objects."""

    linewidth = kwargs.get('linewidth', 5)

    colors = kwargs.get('color',
                        kwargs.get('c',
                                   kwargs.get('colors', None)))

    scale_vect = kwargs.get('dps_scale_vec', 1)

    _, colormap, _ = prepare_colormap(colors,
                                      dotprops=x,
                                      use_neuron_color=kwargs.get('use_neuron_color', False),
                                      color_range=255)

    trace_data = []
    for i, dp in enumerate(x.itertuples()):
        # Get Name
        name = getattr(dp, 'name', getattr(dp, 'gene_name', None))
        c = colormap[i]

        # Prepare lines - this is based on nat:::plot3d.dotprops
        halfvect = dp.points[['x_vec', 'y_vec', 'z_vec']] / 2 * scale_vect

        starts = dp.points[['x', 'y', 'z']].values - halfvect.values
        ends = dp.points[['x', 'y', 'z']].values + halfvect.values

        x_coords = [n for sublist in zip(
            starts[:, 0], ends[:, 0], [None] * starts.shape[0]) for n in sublist]
        y_coords = [n for sublist in zip(
            starts[:, 1], ends[:, 1], [None] * starts.shape[0]) for n in sublist]
        z_coords = [n for sublist in zip(
            starts[:, 2], ends[:, 2], [None] * starts.shape[0]) for n in sublist]

        try:
            c = 'rgb{}'.format(c)
        except BaseException:
            c = 'rgb(10,10,10)'

        trace_data.append(go.Scatter3d(x=x_coords,
                                       y=y_coords,
                                       z=z_coords,
                                       mode='lines',
                                       line=dict(
                                           color=c,
                                           width=linewidth
                                       ),
                                       name=name,
                                       legendgroup=name,
                                       showlegend=True,
                                       hoverinfo='none'
                                       ))

    return trace_data


def volume2plotly(x, **kwargs):
    """ Convert Volumes to plotly objects. """

    colors = kwargs.get('color',
                        kwargs.get('c',
                                   kwargs.get('colors', None)))

    _, _, colormap = prepare_colormap(colors,
                                      volumes=x,
                                      use_neuron_color=kwargs.get('use_neuron_color', False),
                                      color_range=255)

    trace_data = []
    for i, v in enumerate(x):
        # Skip empty data
        if isinstance(v.vertices, np.ndarray):
            if not v.vertices.any():
                continue
        elif not v.vertices:
            continue

        name = getattr(v, 'name', None)

        c = colormap[i]
        if len(c) == 3:
            c = (c[0], c[1], c[2], .5)

        trace_data.append(go.Mesh3d(x=v.vertices[:, 0],
                                    y=v.vertices[:, 1],
                                    z=v.vertices[:, 2],
                                    i=v.faces[:, 0],
                                    j=v.faces[:, 1],
                                    k=v.faces[:, 2],
                                    color=f'rgba{str(c)}',
                                    name=name,
                                    hoverinfo='none'))

    return trace_data


def layout2plotly(**kwargs):
    """ Generate layout for plotly figures."""
    layout = dict(width=kwargs.get('width', 1200),
                  height=kwargs.get('height', 600),
                  autosize=kwargs.get('fig_autosize', False),
                  title=kwargs.get('pl_title', None),
                  scene=dict(xaxis=dict(gridcolor='rgb(255, 255, 255)',
                                        zerolinecolor='rgb(255, 255, 255)',
                                        showbackground=False,
                                        showline=False,
                                        showgrid=False,
                                        backgroundcolor='rgb(240, 240, 240)'
                                        ),
                             yaxis=dict(gridcolor='rgb(255, 255, 255)',
                                        zerolinecolor='rgb(255, 255, 255)',
                                        showbackground=False,
                                        showline=False,
                                        showgrid=False,
                                        backgroundcolor='rgb(240, 240, 240)'
                                        ),
                             zaxis=dict(gridcolor='rgb(255, 255, 255)',
                                        zerolinecolor='rgb(255, 255, 255)',
                                        showbackground=False,
                                        showline=False,
                                        showgrid=False,
                                        backgroundcolor='rgb(240, 240, 240)'
                                        ),
                             camera=dict(up=dict(x=0,
                                                 y=0,
                                                 z=1
                                                 ),
                                         eye=dict(x=-1.7428,
                                                  y=1.0707,
                                                  z=0.7100,
                                                  )
                                         ),
                             aspectratio=dict(x=1, y=1, z=1),
                             aspectmode='data'
                             ),
                  )

    # Need to remove width and height to make autosize actually matter
    if kwargs.get('fig_autosize', False):
        layout.pop('width')
        layout.pop('height')

    return layout



#    This script is part of navis (http://www.github.com/schlegelp/navis).
#    Copyright (C) 2018 Philipp Schlegel
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

import pandas as pd
import numpy as np
import scipy.spatial
import scipy.interpolate

from typing import Union, Optional, List, overload
from typing_extensions import Literal

from .. import config, core, utils

# Set up logging
logger = config.logger

__all__ = ['resample_neuron']


@overload
def resample_neuron(x: 'core.NeuronObject',
                    resample_to: int,
                    inplace: Literal[False],
                    method: str = 'linear',
                    skip_errors: bool = True
                    ) -> 'core.NeuronObject': ...


@overload
def resample_neuron(x: 'core.NeuronObject',
                    resample_to: int,
                    inplace: Literal[True],
                    method: str = 'linear',
                    skip_errors: bool = True
                    ) -> None: ...


@overload
def resample_neuron(x: 'core.NeuronObject',
                    resample_to: int,
                    inplace: bool = False,
                    method: str = 'linear',
                    skip_errors: bool = True
                    ) -> None: ...


def resample_neuron(x: 'core.NeuronObject',
                    resample_to: int,
                    inplace: bool = False,
                    method: str = 'linear',
                    skip_errors: bool = True
                    ) -> Optional['core.NeuronObject']:
    """ Resamples neuron(s) to given resolution.

    Preserves root, leafs, branchpoints. Connectors (if they exist) are mapped
    onto the closest new node.

    Important
    ---------
    This generates an entirely new set of node IDs! Those will be unique
    within a neuron, but you may encounter duplicates across neurons.

    Also: be aware that high-resolution neurons will use A LOT of memory.

    Parameters
    ----------
    x :                 TreeNeuron | NeuronList
                        Neuron(s) to resample.
    resample_to :       int
                        Target sampling resolution, i.e. one node every
                        N units of cable. Note that hitting the exact
                        sampling resolution might not be possible e.g. when
                        a branch is shorter than target resolution.
    method :            str, optional
                        See ``scipy.interpolate.interp1d`` for possible
                        options. By default, we're using linear interpolation.
    inplace :           bool, optional
                        If True, will modify original neuron. If False, a
                        resampled copy is returned.
    skip_errors :       bool, optional
                        If True, will skip errors during interpolation and
                        only print summary.

    Returns
    -------
    TreeNeuron/List
                        Downsampled neuron(s). Only if ``inplace=False``.

    Examples
    --------
    >>> import navis
    >>> n = navis.example_neurons(1)
    >>> # Check sampling resolution (nodes/cable)
    >>> round(n.sampling_resolution, 3)
    191.0
    >>> # Resample to 1 micron (example neurons are in nm space)
    >>> n_rs = navis.resample_neuron(n,
    ...                              resample_to=1000,
    ...                              inplace=False)
    >>> round(n_rs.sampling_resolution)
    962.0

    See Also
    --------
    :func:`navis.downsample_neuron`
                        This function reduces the number of nodes instead of
                        resample to certain resolution. Useful if you are
                        just after some simplification e.g. for speeding up
                        your calculations or you want to preserve node IDs.
    """

    if isinstance(x, core.NeuronList):
        if not inplace:
            x = x.copy()
        results = [resample_neuron(x[i], resample_to,
                                   method=method, inplace=True,
                                   skip_errors=skip_errors)
                   for i in config.trange(x.shape[0],
                                          desc='Resampl. neurons',
                                          disable=config.pbar_hide,
                                          leave=config.pbar_leave)]
        if not inplace:
            return core.NeuronList(results)
        return None
    elif not isinstance(x, core.TreeNeuron):
        raise TypeError(f'Unable to resample data of type "{type(x)}"')

    if not inplace:
        x = x.copy()

    # Collect some information for later
    nodes = x.nodes.set_index('node_id', inplace=False)
    locs = nodes[['x', 'y', 'z']]
    radii = nodes['radius'].to_dict()

    new_nodes: List = []
    max_tn_id = x.nodes.node_id.max() + 1

    errors = 0

    # Iterate over segments
    for i, seg in enumerate(config.tqdm(x.small_segments,
                                        desc='Proc. segments',
                                        disable=config.pbar_hide,
                                        leave=False)):
        # Get coordinates
        coords = locs.loc[seg].values.astype(float)
        # Get radii
        rad = [radii[tn] for tn in seg]

        # Vecs between subsequently measured points
        vecs = np.diff(coords.T)

        # path: cum distance along points (norm from first to ith point)
        path = np.cumsum(np.linalg.norm(vecs, axis=0))
        path = np.insert(path, 0, 0)

        # If path is too short, just keep the first and last node
        if path[-1] < resample_to or (method == 'cubic' and len(seg) <= 3):
            new_nodes += [[seg[0], seg[-1],
                           coords[0][0], coords[0][1], coords[0][2],
                           radii[seg[0]]]]
            continue

        # Coords of interpolation
        n_nodes = int(path[-1] / resample_to)
        interp_coords = np.linspace(path[0], path[-1], n_nodes)

        try:
            sampleX = scipy.interpolate.interp1d(path, coords[:, 0],
                                                 kind=method)
            sampleY = scipy.interpolate.interp1d(path, coords[:, 1],
                                                 kind=method)
            sampleZ = scipy.interpolate.interp1d(path, coords[:, 2],
                                                 kind=method)
            sampleR = scipy.interpolate.interp1d(path, rad,
                                                 kind=method)
        except ValueError as e:
            if skip_errors:
                errors += 1
                new_nodes += x.nodes.loc[x.nodes.node_id.isin(seg[:-1]),
                                         ['node_id', 'parent_id',
                                          'x', 'y', 'z',
                                          'radius']].values.tolist()
                continue
            else:
                raise e

        # Sample each dim
        xnew = sampleX(interp_coords)
        ynew = sampleY(interp_coords)
        znew = sampleZ(interp_coords)
        rnew = sampleR(interp_coords)

        # Generate new coordinates
        new_coords = np.array([xnew, ynew, znew]).T

        # Generate new ids (start and end node IDs of this segment)
        new_ids = seg[:1] + [max_tn_id +
                             i for i in range(len(new_coords) - 2)] + seg[-1:]

        # Keep track of new nodes
        new_nodes += [[tn, pn, co[0], co[1], co[2], r]
                      for tn, pn, co, r in zip(new_ids[:-1],
                                               new_ids[1:],
                                               new_coords,
                                               rnew)]

        # Increase max index
        max_tn_id += len(new_ids)

    if errors:
        logger.warning(f'{errors} ({errors/i:.0%}) segments skipped due to '
                       'errors')

    # Add root node(s)
    root = x.nodes.loc[x.nodes.node_id.isin(utils.make_iterable(x.root)),
                       ['node_id', 'parent_id', 'x', 'y', 'z', 'radius']]
    new_nodes += [list(r) for r in root.values]

    # Generate new nodes dataframe
    new_nodes = pd.DataFrame(data=new_nodes,
                             columns=['node_id', 'parent_id',
                                      'x', 'y', 'z', 'radius'],
                             dtype=object
                             )

    # Convert columns to appropriate dtypes
    dtypes = {'node_id': int, 'parent_id': int, 'x': float, 'y': float,
              'z': float, 'radius': float}

    for k, v in dtypes.items():
        new_nodes[k] = new_nodes[k].astype(v)

    # Remove duplicate nodes (branch points)
    new_nodes = new_nodes[~new_nodes.node_id.duplicated()]

    if x.has_connectors:
        # Map connectors back:
        # 1. Get position of old synapse-bearing nodes
        old_tn_position = x.nodes.set_index('node_id',
                                            inplace=False).loc[x.connectors.node_id,
                                                               ['x', 'y', 'z']].values
        # 2. Get closest neighbours
        distances = scipy.spatial.distance.cdist(old_tn_position,
                                                 new_nodes[['x', 'y', 'z']].values)
        min_ix = np.argmin(distances, axis=1)
        # 3. Map back onto neuron
        x.connectors['node_id'] = new_nodes.iloc[min_ix].node_id.values

    # Set nodes
    x.nodes = new_nodes

    # Clear and regenerate temporary attributes
    x._clear_temp_attr()

    if not inplace:
        return x
    else:
        return None



#    This script is part of navis (http://www.github.com/schlegelp/navis).
#    Copyright (C) 2018 Philipp Schlegel
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.


""" This module contains functions for intersections.
"""

import numpy as np
import pandas as pd

from ..core.volumes import Volume

from scipy.spatial import ConvexHull


def in_volume_convex(points: np.ndarray,
                     volume: Volume,
                     approximate: bool = False,
                     ignore_axis: list = []) -> np.ndarray:
    """ Uses scipy to test if points are within a given volume.
    The idea is to test if adding the point to the pointcloud changes the
    convex hull -> if yes, that point is outside the convex hull.
    """

    verts = volume.vertices

    if not approximate:
        intact_hull = ConvexHull(verts)
        intact_verts = list(intact_hull.vertices)

        if isinstance(points, list):
            points = np.array(points)
        elif isinstance(points, pd.DataFrame):
            points = points.to_matrix()

        return [list(ConvexHull(np.append(verts, list([p]), axis=0)).vertices) == intact_verts for p in points]
    else:
        bbox = [(min([v[0] for v in verts]), max([v[0] for v in verts])),
                (min([v[1] for v in verts]), max([v[1] for v in verts])),
                (min([v[2] for v in verts]), max([v[2] for v in verts]))
                ]

        for a in ignore_axis:
            bbox[a] = (float('-inf'), float('inf'))

        return [False not in [bbox[0][0] < p.x < bbox[0][1], bbox[1][0] < p.y < bbox[1][1], bbox[2][0] < p.z < bbox[2][1], ] for p in points]



#    This script is part of navis (http://www.github.com/schlegelp/navis).
#    Copyright (C) 2018 Philipp Schlegel
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

from .. import config

import pandas as pd

__all__ = ['Dotprops']

# Set up logging
logger = config.logger


class Dotprops(pd.DataFrame):
    """ Class to hold dotprops. This is essentially a pandas DataFrame - we
    just use it to tell dotprops from other objects.

    See Also
    --------
    :func:`navis.interfaces.r.dotprops2py`
        Converts R dotprops to :class:`~navis.Dotprops`.

    Notes
    -----
    This class is still in the making but the idea is to write methods for it
    like .plot3d(), .to_X().
    """
